// TTGO_Tracker v1.00 : Programme TTGO LoRa.
// Ecrit par Frank LOBA.
// Date création : 25/03/2023.
// Dernière mise à jour : 25/03/2023.


/**** INCLUSION DES FICHIERS EN-TETE ****/
#include <SPI.h>
#include <LoRa.h>

#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#include <HardwareSerial.h>
#include <TinyGPSPlus.h>

#include <math.h>


/**** DEFINITION DES CONSTANTES ****/
#define OLED_SDA 21
#define OLED_SCL 22 
#define OLED_RST 16

#define OLED_SCREEN_WIDTH   128
#define OLED_SCREEN_HEIGHT  64

#define OLED_ADRESS 0x3C


#define LORA_SCK  5
#define LORA_SS   18
#define LORA_MISO 19  
#define LORA_MOSI 27
#define LORA_RST  23
#define LORA_DIO0 26

#define LORA_FREQUENCE         433775000L
#define LORA_SPREADING_FACTOR  12
#define LORA_SIGNAL_BANDWIDTH  125000L
#define LORA_CODING_RATE_4     5
#define LORA_SYNC_WORD         0x12

#define LORA_START_BYTE_1 '<'
#define LORA_START_BYTE_2 0xFF
#define LORA_START_BYTE_3 0x01


#define GPS_SERIAL_NUM  1
#define GPS_RX_PIN      34
#define GPS_TX_PIN      12

// Période d'émission en minutes.
#define APRS_PERIODE_TX   10                           
// Distance parcourue en mètres pour émettre.
#define APRS_DISTANCE_TX  1000                        
// Indicatif de l'émetteur.
#define APRS_SOURCE       "F4XXX-12"                  
// Indicatif du destinataire avec le chemin.
#define APRS_DESTINATAIRE "APLG01,WIDE1-1,WIDE2-1"
// Overlay APRS.    
#define APRS_OVERLAY      '/'
// Symbole APRS.
#define APRS_SYMBOLE      '['
// Commentaire de la trame APRS.
#define APRS_COMMENTAIRE  "www.f4hdb.fr"


/**** DECLARATION DES FONCTIONS GLOBALES ****/
String LatitudeAprs(RawDegrees Latitude);
String LongitudeAprs(RawDegrees Longitude);


/**** DEFINITION DES VARIABLES GLOBALES ****/
Adafruit_SSD1306 Display(OLED_SCREEN_WIDTH, OLED_SCREEN_HEIGHT, &Wire, OLED_RST);

// Trame APRS au format TNC2.
String Trame;

HardwareSerial GPSSerial(GPS_SERIAL_NUM);
TinyGPSPlus Gps;

// Ancien instant d'émission.
int InstantTx;

// Distance parcourue.
double Distance;  

// Latitude d'émission.
double LatitudeTx;

// Longitude d'émission.
double LongitudeTx;



/**** DEFINITION DE LA FONCTION D'INITIALISATION ****/
void setup()
{
  // Initialisation de l'interface série (USB).
  Serial.begin(115200);

  // Initialisation de l'écran OLED.
  Wire.begin(OLED_SDA, OLED_SCL);
  
  if(!Display.begin(SSD1306_SWITCHCAPVCC, OLED_ADRESS, false, false))
  {
    // Ecriture d'un message d'erreur dans le port série en cas d'échec.
    Serial.println("Echec initialisation ecran");
    
    while(1);
  }

  // Effacement de l'écran, définition de la couleur et de la taille du texte.
  Display.clearDisplay();
  Display.setTextColor(WHITE);
  Display.setTextSize(1);

  // Affectation des broches pour LoRa.
  SPI.begin(LORA_SCK, LORA_MISO, LORA_MOSI, LORA_SS);
  LoRa.setPins(LORA_SS, LORA_RST, LORA_DIO0);


  // Défintion de la fréquence et démarrage du LoRa.
  if(!LoRa.begin(LORA_FREQUENCE))
  {
    // Ecriture d'un message d'erreur dans le port série en cas d'échec.
    Serial.println("Echec demarrage LoRa");

    // Affichage d'un message à l'écran en cas d'échec.
    Display.setCursor(0,0);
    Display.println("Echec demarrage LoRa");
    Display.display();
    
    while (1);
  }

  // Définition des paramètres système pour LoRa.
  LoRa.setSpreadingFactor(LORA_SPREADING_FACTOR);
  LoRa.setSignalBandwidth(LORA_SIGNAL_BANDWIDTH);
  LoRa.setCodingRate4(LORA_CODING_RATE_4);
  LoRa.enableCrc();
  LoRa.setSyncWord(LORA_SYNC_WORD);

  // Ecriture d'un message dans le port série en cas de succès.
  Serial.println("LoRa OK");

  // Affichage d'un message à l'écran en cas de succès.
  Display.setCursor(0,0);
  Display.println("LoRa OK");
  Display.display();

  // Initialisation du GPS.
  GPSSerial.begin(9600, SERIAL_8N1, GPS_RX_PIN, GPS_TX_PIN);

  // Initialisation de l'instant d'émission.
  InstantTx = millis();

  // Initialisation de la distance parcoure.
  Distance = 0.0;

  // Initialisation de la latitude d'émission.
  LatitudeTx = 0.0;

  // Initialisation de la longitude d'émission.
  LongitudeTx = 0.0;
}


/**** DEFINITION DE LA FONCTION D'EXECUTION ****/
void loop()
{  
  // Détection de la réception d'un paquet.
  int PacketSize = LoRa.parsePacket();
  if (PacketSize)
  {
    // Effacement de l'écran.
    Display.clearDisplay();

    // Positionnement du curseur.
    Display.setCursor(0,0);

    // Affichage de "RX" à l'écran.
    Display.println("**** RX ****");

    // Initialisation de la trame TNC2.
    Trame = "";

    // Détection de l'en-tête LoRa.
    if(LoRa.available() && LoRa.read() == LORA_START_BYTE_1)
    if(LoRa.available() && LoRa.read() == LORA_START_BYTE_2)
    if(LoRa.available() && LoRa.read() == LORA_START_BYTE_3)
    
    // Lecture du paquet au format TNC2.
    while(LoRa.available())
    Trame += (char)LoRa.read();

    // Affichage de la trame au format TNC2 à l'écran.
    Display.println(Trame);
    Display.display();

    // Attente de 5 secondes.
    delay(5000);
  }
  // Si l'on ne reçoit pas de trames, on effectue le traçage.
  else
  {
    while (GPSSerial.available() > 0)
    Gps.encode(GPSSerial.read());

    if(Gps.location.isUpdated() || Gps.time.isUpdated())
    {
      // Calcul de la distance parcourue en mètres.
      // La circonférence de la Terre est de 40000 Km.
      // On utilise le théorème de Pythagore et la trigonométrie.
      Distance = 1000.0 * sqrt(pow((40000.0 / 360.0 * (Gps.location.lat() - LatitudeTx)), 2) + pow((40000.0 * cos(Gps.location.lat()) / 360.0 * (Gps.location.lng() - LongitudeTx)), 2));
      
      // Effacement de l'écran.
      Display.clearDisplay();
  
      // Positionnement du curseur.
      Display.setCursor(0,0);
  
      // Affichage de "GPS" à l'écran.
      Display.println("**** GPS ****");

      Display.print("Latitude : ");
      Display.println(LatitudeAprs(Gps.location.rawLat()));
      Display.print("Longitude : ");
      Display.println(LongitudeAprs(Gps.location.rawLng()));
      Display.print("Satellites: ");
      Display.println(Gps.satellites.value());
      Display.print("Altitude : ");
      Display.print(Gps.altitude.feet() / 3.2808);
      Display.println("m");
      Display.print("Time : ");
      Display.print(Gps.time.hour());
      Display.print(":");
      Display.print(Gps.time.minute());
      Display.print(":");
      Display.println(Gps.time.second());
      Display.print("Distance : ");
      Display.print(Distance);
      Display.println("m");
      Display.display();
    }
  
    if(((Distance > APRS_DISTANCE_TX) || ((millis() - InstantTx) > (APRS_PERIODE_TX * 60000))) && Gps.location.isValid())
    {
      // Mise à jour de l'instant d'émission.
      InstantTx = millis();
      
      // Mise à jour de la latitude et la longitude d'émission.
      LatitudeTx = Gps.location.lat();
      LongitudeTx = Gps.location.lng();

      // Constitution de la trame APRS.
      Trame = APRS_SOURCE;
      Trame += '>';
      Trame += APRS_DESTINATAIRE;
      Trame += ":=";
      Trame += LatitudeAprs(Gps.location.rawLat());
      Trame += APRS_OVERLAY;
      Trame += LongitudeAprs(Gps.location.rawLng());
      Trame += APRS_SYMBOLE;
      Trame += APRS_COMMENTAIRE;
      
      // Effacement de l'écran.
      Display.clearDisplay();

      // Positionnement du curseur.
      Display.setCursor(0,0);

      // Affichage de "TX" à l'écran.
      Display.println("**** TX ****");

      // Commencement de l'émission LoRa.
      LoRa.beginPacket();
            
      // Emission de l'en-tête LoRa.
      LoRa.write(LORA_START_BYTE_1);
      LoRa.write(LORA_START_BYTE_2);
      LoRa.write(LORA_START_BYTE_3);
            
      // Emission de la trame TNC2 en LoRa.
      LoRa.write((const uint8_t *)Trame.c_str(), Trame.length());
            
      // Fin de l'émission LoRa.
      LoRa.endPacket();

      // Affichage de la trame au format TNC2 à l'écran.
      Display.println(Trame);     
      Display.display();

      // Attente de 5 secondes.
      delay(5000);
    }
  }

  // Attente de 20 milli secondes.
  delay(20);
}


/**** DEFINITION DE LA FONCTION DE CONVERSION DE LA LATITUDE AU FORMAT APRS ****/
String LatitudeAprs(RawDegrees Latitude)
{
  char Tampon[20];
  
  sprintf(Tampon, "%02d%05.2f", Latitude.deg, Latitude.billionths / 1000000000.0 * 60.0);
  
  String Chaine(Tampon);

  if(Latitude.negative) Chaine += 'S';
  else Chaine += 'N';
  
  return Chaine;
}



/**** DEFINITION DE LA FONCTION DE CONVERSION DE LA LONGITUDE AU FORMAT APRS ****/
String LongitudeAprs(RawDegrees Longitude)
{
  char Tampon[20];

  sprintf(Tampon, "%03d%05.2f", Longitude.deg, Longitude.billionths / 1000000000.0 * 60.0);

  String Chaine(Tampon);

  if(Longitude.negative) Chaine += 'W';
  else Chaine += 'E';
  
  return Chaine;
}
